# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

from ._codec_audio_copy import AudioCopy
from ._codec_audio_none import NoAudio
from ._codec_avc import Avc
from ._codec_avc_preset import VcPreset
from ._codec_hevc import Hevc, VcPreset, HevcBitrateSpecifiedForLosslessError, \
    HevcBitrateNotSpecifiedError, HevcLosslessAndNearLosslessError
from ._codec_prores_ks import Prores, ProresProfile
from ._codec_video_copy import VideoCopy
from ._command import FfmpegCommand, VideoCodecNotSpecifiedError, \
    AudioCodecNotSpecifiedError
from ._common import Scale
from ._filter_crop import Crop
from ._filter_pad import Pad
from ._filter_transpose import Transpose
from ._math_cropping import crop_and_scale
from ._math_padding import _letterbox
from ._pf_10_pixfmts_stdout_parser import pixfmt_alpha
from ._pf_15_pixfmt_subsampling import pixfmt_subsampling
from ._pf_20_pixfmt_bpc import pixfmt_bpc




# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

from typing import NamedTuple, Union, Dict, Iterable, Tuple


class Scale(NamedTuple):
    width: int
    height: int
    downscale_only: bool = False


def colon_splitted_pairs(
        items: Union[Dict[str, str], Iterable[Tuple[str, str]]]) -> str:
    if isinstance(items, dict):
        items = items.items()
    return ':'.join(f"{k}={v}" for (k, v) in items)


def frame_dimension_spec(iw_or_ih: str, value: int,
                         downscale_only: bool) -> str:
    assert iw_or_ih in ('iw', 'ih')
    if not downscale_only:
        return str(value)

    # итак, мы собираемся только уменьшать. Значит, хотим что-то вроде
    #   'min(iw,1920)':'min(ih,1080)' (если оба размера заданы)
    # или
    #   -2:'min(ih,1080)' (если один задан, а второй должен быть пропорционален)

    if value >= 0:
        return f"'min({iw_or_ih},{value})'"
    else:
        assert value < 0
        return str(value)



# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

import shlex
from typing import List, Iterable, Tuple, Optional


class ArgsSubset:
    # хранит аргументы в виде List[str].
    # Позволяет задавать/читать их и иначе: в виде строки, в виде списка пар
    def __init__(self):
        self.list: List[str] = list()

    @property
    def string(self) -> str:
        return shlex.join(self.list)

    @string.setter
    def string(self, value: str):
        self.list = shlex.split(value)

    def pairs(self) -> Iterable[Tuple[str, Optional[str]]]:

        def is_key(s):
            return s.startswith('-')

        prev_key = None
        for arg in self.list:
            if prev_key is None:
                if is_key(arg):
                    prev_key = arg
                    continue
                else:
                    raise ValueError(arg)
            else:
                assert prev_key is not None
                if is_key(arg):
                    yield prev_key, None
                    prev_key = arg
                else:
                    assert prev_key is not None and not is_key(arg)
                    yield prev_key, arg
                    prev_key = None

        if prev_key is not None:
            yield prev_key, None



# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

"""
It would probably be easier and more reliable to create a table with all pixel
formats. But I was too lazy. Therefore, below is a heuristic parser of the data
printed by the command "ffmpeg -pix_fmts".

The `_pix_format_to_depth` dictionary allows you to overload the parser
results with constants.

================================================================================

ffmpeg -pix_fmts

Pixel formats:
I.... = Supported Input  format for conversion
.O... = Supported Output format for conversion
..H.. = Hardware accelerated format
...P. = Paletted format
....B = Bitstream format
FLAGS NAME            NB_COMPONENTS BITS_PER_PIXEL
-----
IO... yuv420p                3            12
IO... yuyv422                3            16
IO... rgb24                  3            24
IO... bgr24                  3            24
IO... yuv422p                3            16
IO... yuv444p                3            24
IO... yuv410p                3             9
IO... yuv411p                3            12
IO... gray                   1             8
IO..B monow                  1             1
IO..B monob                  1             1
I..P. pal8                   1             8
IO... yuvj420p               3            12
IO... yuvj422p               3            16
IO... yuvj444p               3            24
IO... uyvy422                3            16
..... uyyvyy411              3            12
IO... bgr8                   3             8
.O..B bgr4                   3             4
IO... bgr4_byte              3             4
IO... rgb8                   3             8
.O..B rgb4                   3             4
IO... rgb4_byte              3             4
IO... nv12                   3            12
IO... nv21                   3            12
IO... argb                   4            32
IO... rgba                   4            32
IO... abgr                   4            32
IO... bgra                   4            32
IO... gray16be               1            16
IO... gray16le               1            16
IO... yuv440p                3            16
IO... yuvj440p               3            16
IO... yuva420p               4            20
IO... rgb48be                3            48
IO... rgb48le                3            48
IO... rgb565be               3            16
IO... rgb565le               3            16
IO... rgb555be               3            15
IO... rgb555le               3            15
IO... bgr565be               3            16
IO... bgr565le               3            16
IO... bgr555be               3            15
IO... bgr555le               3            15
..H.. vaapi_moco             0             0
..H.. vaapi_idct             0             0
..H.. vaapi_vld              0             0
IO... yuv420p16le            3            24
IO... yuv420p16be            3            24
IO... yuv422p16le            3            32
IO... yuv422p16be            3            32
IO... yuv444p16le            3            48
IO... yuv444p16be            3            48
..H.. dxva2_vld              0             0
IO... rgb444le               3            12
IO... rgb444be               3            12
IO... bgr444le               3            12
IO... bgr444be               3            12
IO... ya8                    2            16
IO... bgr48be                3            48
IO... bgr48le                3            48
IO... yuv420p9be             3            13
IO... yuv420p9le             3            13
IO... yuv420p10be            3            15
IO... yuv420p10le            3            15
IO... yuv422p10be            3            20
IO... yuv422p10le            3            20
IO... yuv444p9be             3            27
IO... yuv444p9le             3            27
IO... yuv444p10be            3            30
IO... yuv444p10le            3            30
IO... yuv422p9be             3            18
IO... yuv422p9le             3            18
IO... gbrp                   3            24
IO... gbrp9be                3            27
IO... gbrp9le                3            27
IO... gbrp10be               3            30
IO... gbrp10le               3            30
IO... gbrp16be               3            48
IO... gbrp16le               3            48
IO... yuva422p               4            24
IO... yuva444p               4            32
IO... yuva420p9be            4            22
IO... yuva420p9le            4            22
IO... yuva422p9be            4            27
IO... yuva422p9le            4            27
IO... yuva444p9be            4            36
IO... yuva444p9le            4            36
IO... yuva420p10be           4            25
IO... yuva420p10le           4            25
IO... yuva422p10be           4            30
IO... yuva422p10le           4            30
IO... yuva444p10be           4            40
IO... yuva444p10le           4            40
IO... yuva420p16be           4            40
IO... yuva420p16le           4            40
IO... yuva422p16be           4            48
IO... yuva422p16le           4            48
IO... yuva444p16be           4            64
IO... yuva444p16le           4            64
..H.. vdpau                  0             0
IO... xyz12le                3            36
IO... xyz12be                3            36
..... nv16                   3            16
..... nv20le                 3            20
..... nv20be                 3            20
IO... rgba64be               4            64
IO... rgba64le               4            64
IO... bgra64be               4            64
IO... bgra64le               4            64
IO... yvyu422                3            16
IO... ya16be                 2            32
IO... ya16le                 2            32
IO... gbrap                  4            32
IO... gbrap16be              4            64
IO... gbrap16le              4            64
..H.. qsv                    0             0
..H.. mmal                   0             0
..H.. d3d11va_vld            0             0
..H.. cuda                   0             0
IO... 0rgb                   3            24
IO... rgb0                   3            24
IO... 0bgr                   3            24
IO... bgr0                   3            24
IO... yuv420p12be            3            18
IO... yuv420p12le            3            18
IO... yuv420p14be            3            21
IO... yuv420p14le            3            21
IO... yuv422p12be            3            24
IO... yuv422p12le            3            24
IO... yuv422p14be            3            28
IO... yuv422p14le            3            28
IO... yuv444p12be            3            36
IO... yuv444p12le            3            36
IO... yuv444p14be            3            42
IO... yuv444p14le            3            42
IO... gbrp12be               3            36
IO... gbrp12le               3            36
IO... gbrp14be               3            42
IO... gbrp14le               3            42
IO... yuvj411p               3            12
I.... bayer_bggr8            3             8
I.... bayer_rggb8            3             8
I.... bayer_gbrg8            3             8
I.... bayer_grbg8            3             8
I.... bayer_bggr16le         3            16
I.... bayer_bggr16be         3            16
I.... bayer_rggb16le         3            16
I.... bayer_rggb16be         3            16
I.... bayer_gbrg16le         3            16
I.... bayer_gbrg16be         3            16
I.... bayer_grbg16le         3            16
I.... bayer_grbg16be         3            16
..H.. xvmc                   0             0
IO... yuv440p10le            3            20
IO... yuv440p10be            3            20
IO... yuv440p12le            3            24
IO... yuv440p12be            3            24
IO... ayuv64le               4            64
..... ayuv64be               4            64
..H.. videotoolbox_vld       0             0
IO... p010le                 3            15
IO... p010be                 3            15
IO... gbrap12be              4            48
IO... gbrap12le              4            48
IO... gbrap10be              4            40
IO... gbrap10le              4            40
..H.. mediacodec             0             0
IO... gray12be               1            12
IO... gray12le               1            12
IO... gray10be               1            10
IO... gray10le               1            10
IO... p016le                 3            24
IO... p016be                 3            24
..H.. d3d11                  0             0
IO... gray9be                1             9
IO... gray9le                1             9
IO... gbrpf32be              3            96
IO... gbrpf32le              3            96
IO... gbrapf32be             4            128
IO... gbrapf32le             4            128
..H.. drm_prime              0             0
..H.. opencl                 0             0
IO... gray14be               1            14
IO... gray14le               1            14
IO... grayf32be              1            32
IO... grayf32le              1            32
IO... yuva422p12be           4            36
IO... yuva422p12le           4            36
IO... yuva444p12be           4            48
IO... yuva444p12le           4            48
IO... nv24                   3            24
IO... nv42                   3            24
..H.. vulkan                 0             0
..... y210be                 3            20
I.... y210le                 3            20
IO... x2rgb10le              3            30
..... x2rgb10be              3            30

"""

import re
import warnings
from typing import Optional

from vtcff._pf_10_pixfmts_stdout_parser import pixfmt_to_spec
from vtcff._pf_15_pixfmt_subsampling import pixfmt_subsampling


def _int_if_round(f: float) -> Optional[int]:
    result = int(f)
    if abs(result - f) < 0.0000001:
        return result
    else:
        return None


def _ending(text: str) -> Optional[int]:
    m = re.search(r'(\d+)(?:le|be|$)', text)
    if m:
        return int(m.group(1))
    return None


_pix_format_to_depth = {
    # todo add important formats that are not recognized by the heuristics
    'yuv410p': 8
}


def _subsampling_factor(name: str, components: int) -> Optional[float]:
    # 4:4:4 (no alpha)   -> (4+4+4)/(4+4+4)     = 1
    # 4:2:2 (no alpha)   -> (4+2+2)/(4+4+4)     = 2/3
    # 4:4:4 (with alpha) -> (4+4+4+4)/(4+4+4+4) = 1
    # 4:2:2 (with alpha) -> (4+4+2+2)/(4+4+4+4) = 3/4

    if components not in (3, 4):
        return None
    subsampling_str = pixfmt_subsampling(name)
    if not subsampling_str:
        return None
    assert len(subsampling_str) == 3 and subsampling_str.isdigit()
    pixels_total = sum(int(c) for c in subsampling_str)
    assert 1 <= pixels_total <= 12
    pixels_max = 12
    if components == 4:
        pixels_total += 4  # assuming alpha channel always have all 4 pixels
        pixels_max += 4
    else:
        assert components == 3
    return pixels_total / pixels_max


def _guess_from_subsampling(name: str,
                            components: int,
                            bits_per_pixel: int) -> Optional[int]:
    f = _subsampling_factor(name, components)
    if f:
        return _int_if_round(bits_per_pixel / components / f)
    else:
        return None


def _guess_from_suffix(name: str, components: int, bits_per_pixel: int) -> \
        Optional[int]:
    e: Optional[int] = _ending(name)
    if e is not None:
        if e == bits_per_pixel:
            # like in rgb24, 24 means 8+8+8, 8 bit per channel
            return _int_if_round(bits_per_pixel / components)
        f = _subsampling_factor(name, components) or 1.0
        if e * components * f == bits_per_pixel:
            # yuv444p12le: 12 bits per channel
            # yuv440p10be: 10 bits per channel
            return e
    return None


def _guess_bpc(name: str, components: int, bits_per_pixel: int) \
        -> Optional[int]:
    result = (_pix_format_to_depth.get(name)
              or _guess_from_suffix(name, components, bits_per_pixel)
              or _guess_from_subsampling(name, components, bits_per_pixel))
    if result is None:
        warnings.warn(f"Color depth of {name} format is unknown.")
    return result


def pixfmt_bpc(pixfmt: str, ffmpeg_exe: str) -> Optional[int]:
    """Returns the estimated number of bits-per-color-channel after decoding
    from the specified pixel format. This value optimistically indicates the
    maximum number of color gradations in each of the channels R, G, B, A."""
    # todo test alternate path to ffmpeg_exe
    spec = pixfmt_to_spec(pixfmt, ffmpeg_exe)
    return _guess_bpc(spec.name, spec.nb_components, spec.bits_per_pixel)



# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

from typing import Iterable, Optional
from typing import Tuple

from ._codec import AudioCodec


class NoAudio(AudioCodec):
    def args(self) -> Iterable[Tuple[str, Optional[str]]]:
        yield "-an", None



# SPDX-FileCopyrightText: (c) 2016-2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

from typing import Tuple

from vtcff._filter_crop import Crop
from ._command import FfmpegCommand, Scale


def crop_height(width, aspect: Tuple[int, int], div=1):
    # print("crop_height %dxX->%d:%d"%(width,aspectPair[0],aspectPair[1]))

    # aspectPair - это FrameSize, обозначающий желаемое соотношение сторон:
    # (1920,1080) или просто (16,9).
    # width - фиксированная ширина будущего кадра.
    # Метод подбирает соответствующую такой ширине высоту.
    #
    # Сложность в том, что ширина и высота у нас по определению числа целые,
    # а aspect - дробное. Поэтому примерно мы ищем такой результат:
    #   w/h = aspect
    #   w/aspect = round(h)
    #   h*aspect = round(w)
    #
    # Но и это не всегда возможно. Например, целевые пропорции это 4096x2160.
    # Фиксированная ширина кадра 5472. Дробное значение соответствующей высоты
    # - 2885.625. В зависимости способа от округления высоты, заданной
    # пропорции будут соответствовать два разрешения:
    #   5470.81 x 2885
    #   5472.71 x 2886
    # Тут очевидно, что изменилась ширина, как ее ни округляй.
    #
    # Поэтому мы подбираем не только высоту, но и ширину уменьшаем до тех пор,
    # пока на найдем пару WxH, соответствующую требованиям.

    aspect_f = aspect[0] / aspect[1]

    for w in range(width, 0, -1):

        if w % div != 0:
            continue

        h = int(round(w / aspect_f))

        if h % div != 0:
            continue

        if round(h * aspect_f) == w:
            return w, h

    return None, None


def crop_width(height, aspect: Tuple[int, int], div=1):
    # меняем местами ширину и высоту
    rotatedAspect = aspect[1], aspect[0]
    rotatedWidth = height

    # находим размер
    rotatedResult = crop_height(rotatedWidth, rotatedAspect,
                                div=div)

    # снова меняем местами
    return rotatedResult[1], rotatedResult[0]


def crop_size(src_size: Tuple[int, int], dst_aspect: Tuple[int, int], div=1):
    # targetAspectPair - два числа, сообщающие о желаемых пропорциях размера кадра.
    # sourceSize - исходный размер кадра.
    # Возвращается максимальный по размер кропнутый кадр нужных пропорций.

    sourceAspect = src_size[0] / src_size[1]
    targetAspect = dst_aspect[0] / dst_aspect[1]

    sourceWidth = src_size[0]
    sourceHeight = src_size[1]

    if sourceAspect > targetAspect:
        # нужно обрезать края слева и справа
        result = crop_width(sourceHeight, dst_aspect,
                            div=div)
    elif sourceAspect < targetAspect:
        # нужно обрезать края сверху и снизу
        result = crop_height(sourceWidth, dst_aspect,
                             div=div)
    else:
        # пропорции и так соответствуют
        result = src_size

    assert result[0] <= src_size[0] and result[1] <= src_size[1]
    return result


def whwh_to_xywh(sourceWidth, sourceHeight,
                 targetWidth, targetHeight,
                 alignX="M", alignY="M"):
    # Два этапа:
    #   1. Отрезаем от исходника лишние пиксели, сделав пропорции
    #   такими, как надо.
    #   2. Пропорционально уменьшаем оставшийся прямоугольник.

    croppedSize = crop_size((sourceWidth, sourceHeight),
                            (targetWidth, targetHeight), div=1)
    # в этом вызове функции мы использовали (targetWidth,targetHeight)
    # как целевые пропорции, а не целевое разрешение. Функция вернула что-то
    # по возможности близкое к первому аргументу, а не ко второму.

    widthDelta = sourceWidth - croppedSize[0]
    heightDelta = sourceHeight - croppedSize[1]

    assert widthDelta >= 0
    assert heightDelta >= 0

    cropTop = 0
    cropBottom = 0
    cropLeft = 0
    cropRight = 0

    if widthDelta > 0:
        if alignX == "M":
            halfWidthDelta = int(round(widthDelta / 2))
            cropLeft = halfWidthDelta
            cropRight = widthDelta - cropLeft
        elif alignX == "L":
            cropRight = widthDelta
            cropLeft = 0
        elif alignX == "R":
            cropLeft = widthDelta
            cropRight = 0
        else:
            raise Exception("Unexpected alignX value: %s." % alignX)

    if heightDelta > 0:
        if alignY == "M":
            halfHeightDelta = int(round(heightDelta / 2))
            cropTop = halfHeightDelta
            cropBottom = heightDelta - cropTop
        elif alignY == "T":
            cropBottom = heightDelta
            cropTop = 0
        elif alignY == "B":
            cropTop = heightDelta
            cropBottom = 0
        else:
            raise Exception("Unexpected alignY value: %s." % alignY)

    croppedWidth = sourceWidth - cropLeft - cropRight
    croppedHeight = sourceHeight - cropTop - cropBottom

    # croppedHeight+=1
    # tmpAspect = croppedWidth / croppedHeight
    # logger.info(
    #     "sourceWidth=%d, cropLeft=%d, cropRight=%d, croppedWidth=%d" % (
    #         sourceWidth, cropLeft, cropRight, croppedWidth))
    # logger.info(
    #     "sourceHeight=%d, cropTop=%d, cropBottom=%d, croppedHeight=%d" % (
    #         sourceHeight, cropTop, cropBottom, croppedHeight))
    # logger.info("croppedWidth/croppedHeight=%d/%d=%.5f" % (
    #     croppedWidth, croppedHeight, croppedWidth / croppedHeight))
    # logger.info("finalHeight %d => finalWidth %.4f" % (
    #     targetHeight, targetHeight * tmpAspect))

    # баг-2016? При размере кропнутой картинки 5471:2885 и scale=-1:2160 ffmpeg
    # почему-то вычислял ширину целевого кадра как 4095, хотя арифметически
    # ширина должна была получиться 2160*(5471/2885) = 4096.1386. Это
    # выяснилось для отдельного видео, которое из размера 5472x3076 я
    # хотел сконвертить в 4096x2160.
    #
    # Чтобы исправить проблему, я просто заменил scale=-1:2160 на
    # scale=-2:2160. Как именно ffmpeg станет это округлять - не проверил

    assert round((sourceWidth - cropLeft - cropRight) / (
            sourceHeight - cropTop - cropBottom)) == round(
        targetWidth / targetHeight)

    return cropLeft, cropTop, croppedWidth, croppedHeight


def crop_and_scale(cmd: FfmpegCommand,
                   src_width, src_height,
                   dst_width, dst_height,
                   align_x="M", align_y="M"):
    x, y, w, h = whwh_to_xywh(sourceWidth=src_width,
                              sourceHeight=src_height,
                              targetWidth=dst_width,
                              targetHeight=dst_height, alignX=align_x,
                              alignY=align_y)
    cmd.crop = Crop(left=x, top=y, width=w, height=h)
    cmd.scale = Scale(-2, dst_height)
    cmd._crop_before_scale()



# SPDX-FileCopyrightText: (c) 2016-2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

import subprocess
import sys
from typing import Iterable, List, NamedTuple, Optional, Dict


def _pix_fmts_lines(ffmpeg_exe: str) -> Iterable[str]:
    result = subprocess.run([ffmpeg_exe, "-pix_fmts"],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            encoding=sys.stdout.encoding or "utf-8")
    if result.returncode != 0:
        raise subprocess.CalledProcessError

    txt = result.stdout

    # txt is something like
    #
    # Pixel formats:
    # I.... = Supported Input  format for conversion
    # .O... = Supported Output format for conversion
    # ..H.. = Hardware accelerated format
    # ...P. = Paletted format
    # ....B = Bitstream format
    # FLAGS NAME            NB_COMPONENTS BITS_PER_PIXEL
    # -----
    # IO... yuv420p                3            12
    # IO... yuyv422                3            16
    # IO... rgb24                  3            24
    # IO... bgr24                  3            24
    # IO... yuv422p                3            16

    started = False

    returned = 0

    for line in txt.splitlines():

        if started:
            returned += 1
            yield line
        elif line == "-----":
            started = True

    if returned <= 10:
        raise ValueError(f"Failed to split output to lines. Output: {txt}")


class PixfmtOutputSpec(NamedTuple):
    flags: str
    name: str
    nb_components: int
    bits_per_pixel: int


def _pix_fmts_tuples(ffmpeg_exe: str) -> List[PixfmtOutputSpec]:
    result = list()
    for line in _pix_fmts_lines(ffmpeg_exe):
        words = line.split()
        pfo = PixfmtOutputSpec(words[0], words[1], int(words[2]),
                               int(words[3]))
        result.append(pfo)
    return result


_pix_format_spec: Optional[Dict[str, PixfmtOutputSpec]] = None


def pixfmt_to_spec(name: str, ffmpeg_exe: str) -> PixfmtOutputSpec:
    """Returns a particular parsed line of `ffmpeg -pix_fmts`"""
    global _pix_format_spec
    if _pix_format_spec is None:
        _pix_format_spec = dict()
        for t in _pix_fmts_tuples(ffmpeg_exe=ffmpeg_exe):
            _pix_format_spec[t.name] = t
    assert _pix_format_spec is not None
    return _pix_format_spec[name]


def pixfmt_alpha(pixfmt: str, ffmpeg_exe: str) -> Optional[bool]:
    spec = pixfmt_to_spec(pixfmt, ffmpeg_exe=ffmpeg_exe)
    if spec.nb_components == 4:
        return True
    elif spec.nb_components == 3:
        return False
    else:
        return None



# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

from typing import Optional, Iterable, Tuple

from ._codec import VideoCodec
from ._codec_avc_preset import VcPreset


class Avc(VideoCodec):
    def __init__(self, preset: VcPreset = None):
        self.preset: Optional[VcPreset] = preset

    def args(self) -> Iterable[Tuple[str, str]]:
        yield "-codec:v", "libx264"

        if self.preset is not None:
            yield "-preset", str(self.preset.value)



# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

from typing import Iterable
from typing import Tuple

from ._codec import AudioCodec


class AudioCopy(AudioCodec):
    def args(self) -> Iterable[Tuple[str, str]]:
        yield "-acodec", "copy"



# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

from enum import unique, IntEnum
from typing import Optional, Tuple, Iterable

from ._codec import VideoCodec


@unique
class ProresProfile(IntEnum):
    PROXY = 0
    """ProRes 422 Proxy"""

    LT = 1
    """ProRes 422 LT"""

    NORMAL = 2
    """ProRes 422"""

    HQ = 3
    """ProRes 422 HQ"""

    FOUR = 4
    """ProRes 4444"""

    XQ = 5
    """ProRes 4444 XQ"""


class Prores(VideoCodec):
    def __init__(self,
                 profile: ProresProfile = ProresProfile.NORMAL,
                 qscale: Optional[int] = None,
                 spoof_vendor: bool = False):
        self.profile: ProresProfile = profile
        self.qscale: Optional[int] = qscale
        self.spoof_vendor: bool = spoof_vendor

    def args(self) -> Iterable[Tuple[str, str]]:
        yield '-codec:v', 'prores_ks'
        if self.spoof_vendor:
            yield '-vendor', 'apl0'
        if self.qscale is not None:
            if self.qscale is not None and not 0 <= self.qscale <= 64:
                raise ValueError(self.qscale)
            yield '-q:v', str(self.qscale)
        if self.profile is not None:
            yield '-profile:v', str(self.profile.value)



from __future__ import annotations

from enum import Enum
from typing import Union


class ColorSpace(Enum):
    # todo use in command
    bt709 = "bt709"

    # ffmpeg bt470bg is BT.470BG or BT.601-6 625
    # (http://underpop.online.fr/f/ffmpeg/help/colorspace.htm.gz)
    bt470bg = "bt470bg"
    """BT.470BG or BT.601-6 625"""

    # ffmpeg smpte170m is SMPTE-170M or BT.601-6 525
    # (http://underpop.online.fr/f/ffmpeg/help/colorspace.htm.gz)
    smpte170m = "smpte170m",
    """BT.601-6 525"""

    # ffmpeg bt2020ncl is BT.2020 with non-constant luminance
    # (http://underpop.online.fr/f/ffmpeg/help/colorspace.htm.gz)
    bt2020ncl = 'bt2020ncl'
    """BT.2020 with non-constant luminance"""

    # not tested yet (not sure how to implement)
    # bt2020 - BT.2020
    # bt2020-10 - BT.2020 for 10-bits content
    # bt2020-12 - BT.2020 for 12-bits content


def safe_str_to_colorspace(s: str) -> Union[ColorSpace, str]:
    try:
        return ColorSpace(s)
    except ValueError:
        return s


def safe_colorspace_to_str(cs: Union[ColorSpace, str]) -> str:
    if isinstance(cs, ColorSpace):
        return cs.value
    else:
        return str(cs)



# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

from typing import Optional, Iterable, Tuple


class FilterBase:
    def __init__(self):
        self._pairs = dict()

    def _set_or_remove(self, key: str, val: Optional[str]):
        if val is not None:
            self._pairs[key] = val
        else:
            assert val is None
            if key in self._pairs:
                del self._pairs[key]
        assert self._pairs.get(key) == val

    @classmethod
    def _to_string(cls, name: str, kv_pairs: Iterable[Tuple[str, str]]):
        return f'{name}={":".join(k + "=" + v for (k, v) in kv_pairs)}'



# SPDX-FileCopyrightText: (c) 2021 Artёm IG <github.com/rtmigo>
# SPDX-License-Identifier: MIT

from enum import IntEnum, unique


@unique
class Transpose(IntEnum):
    # https://stackoverflow.com/questions/3937387/rotating-videos-with-ffmpeg
    COUNTER_CLOCKWISE_VFLIP = 0
    CLOCKWISE = 1
    COUNTER_CLOCKWISE = 2
    CLOCKWISE_VFLIP = 3


class TransposeFilter:
    def __init__(self):
        self.kind = Transpose.CLOCKWISE

    def __str__(self):
        return f"transpose={self.kind.value}"
